/*
 *	Convertion to AMXMod of Lerk Lift by [WHO]Them for metamod
 *
 *	Lerks can carry gorges by walking up to them and pressing use.
 *
 *
 *	Note:		I pretty much copied [WHO]Them's design, fixing bugs where I saw them:
 *				O Gorges no longer take fall damage so they cannot die from lerk dive bombing
 *				O Gorges are carried 16 units below the Lerk so that both players can see easier
 *				O When stuck in the ground the Gorges energy will increase as normal
 *				O Loads more bug fixes, I forget what they were. :o
 *			I added a few extra features:
 *				O Lerks & Gorges are notified who they have picked up, released or been picked up by.
 *				O Players can say /lloff to stop Lerks from picking them up. (/llon to enable)
 *				O Lerk Lift can be disabled by admins with the command "amx_lerklift off".
 *
 *	Commands:	say /lloff		Disables Lerks from picking you up. This setting is stored for all servers.
 *			say /llon		Enable Lerks picking you up.
 *			amx_lerklift off	Turn off Lerk Lift completely. (ADMIN_BAN required)
 *			amx_lerklift on		Turn on Lerk Lift. (ADMIN_BAN required)
 *
 *	Cvars:		None
 *    
 *	Requires:	ns2amx 1.0
 *
 *	Author:		Cheesy Peteza  
 *	Date:		12-May-2004
 *	Email:		cheesy@yoclan.com
 *	irc:		#yo-clan (QuakeNet)
 *
 */

#include <amxmod>
#include <amxmisc>
#include <ns2amx>

#define ALIEN		2
#define IN_USE		(1 << 5)
#define	SOLID_NOT	0		// no interaction with other objects
#define	SOLID_SLIDEBOX	3		// touch on edge, but not an onground
#define FL_DUCKING	(1<<14)		// Player flag -- Player is fully crouched
#define START_DISTANCE	32		// The first search distance for finding a free location in the map
#define MAX_ATTEMPTS	128		// How many times to search in an area for a free space
#define CP_LLNOLIFT	1		// Bitmask stored in user key "cp". I use the same key for all my plugins,
					// so don't just go making up new masks if you edit this plugin!

new g_MaxPlayers
new bool:g_lldisabled
new g_lifting[33], g_lifted[33]
new Float:g_LastUseTime[33]
new Float:g_LastEnergyCheck[33], Float:g_LastEnergy[33]

public plugin_init() {
	register_plugin("AMX Lerk Lift", "1.2", "Cheesy Peteza")
	register_cvar("amx_lerklift_version", "1.2", FCVAR_SERVER)

	register_cvar("lerklift_enabled", "1")

	register_clcmd("say /lloff", "disableLL", 0, "Saying this will stop lerks from picking you up")
	register_clcmd("say_team /lloff", "disableLL", 0, "Saying this will stop lerks from picking you up")
	register_clcmd("say /llon", "enableLL", 0, "Saying this will stop lerks from picking you up")
	register_clcmd("say_team /llon", "enableLL", 0, "Saying this will stop lerks from picking you up")

	register_concmd("amx_lerklift", "switchOnOff", ADMIN_BAN, "<on|off> Enable or disable Lerk Lift")

	g_MaxPlayers = get_maxplayers()

	if (!get_cvar_num("lerklift_enabled")) g_lldisabled = true
}

public server_frame() {
	if (g_lldisabled) return PLUGIN_CONTINUE

	new class

	for (new id; id <= g_MaxPlayers; ++id) {
		if (!is_entity(id)) continue

		class = get_class(id)

		if (class == CLASS_LERK) {
			if (pev(id, pev_button) & IN_USE) {
				if ( (get_gametime() - g_LastUseTime[id]) < 1.0 ) continue	// Only react to one use
				g_LastUseTime[id] = get_gametime()				// button press a second

				if (g_lifting[id]) {
					freeGorge(g_lifting[id])
				} else {
					new entity, part
					get_user_aiming(id, entity, part)

					if (get_class(entity) == CLASS_GORGE) {
						if (get_range(id, entity) < 200) {
							new CPstring[32]
							get_user_info(entity, "cp", CPstring, 31)
							new CPvalue = str_to_num(CPstring)

							if (CPvalue & CP_LLNOLIFT) {
								client_print(id, print_center, "The Gorge has disabled lifting")
								continue
							}

							g_lifting[id] = entity			//
							g_lifted[entity] = id			// Pick up the Gorge
							set_pev(entity, pev_solid, SOLID_NOT)	//
							notifyStatus(entity, id, 1)		//
						}
					}
				}
			}
		} else if (g_lifted[id]) {	// A player being carried
			if (pev(id, pev_button) & IN_USE) {					// If the carried gorge
				if ( (get_gametime() - g_LastUseTime[id]) < 1.0 ) continue	// presses use,
				g_LastUseTime[id] = get_gametime()				// free it

				freeGorge(id)
				continue
			}

			if ( (class != CLASS_GORGE) ||		// occassional bug here, sometimes gorge has gestating class, weird. :/
			     (get_class(g_lifted[id]) != CLASS_LERK) ||
			     (pev(id, pev_team) != pev(g_lifted[id], pev_team)) ||		// If the gorge has changed
			     (!is_user_alive(id)) ||						// team, class, died or the
			     (!is_user_alive(g_lifted[id])) ) {					// Lerk has died,
				freeGorge(id)							// free it.
				continue
			}

			// Manually increase energy for gorge if its not going up itself
			if ((get_gametime() - g_LastEnergyCheck[id]) >= 0.1 ) {
				g_LastEnergyCheck[id] = get_gametime()				// Increase the energy of a
												// Gorge manually if it is
				new Float:energy = pev_f(id, pev_fuser3)			// stuck in the ground.
				if ( (energy == g_LastEnergy[id]) && (energy != 1000.0) ) {
					energy+=7.8
					if (energy > 1000.0) energy = 1000.0
					set_pev(id, pev_fuser3, energy)
				}
				g_LastEnergy[id] = energy
			}

			// The origin must be set every server frame, otherwise the Gorge will have a laggy view.
			set_pev(id, pev_origin_x, pev_f(g_lifted[id], pev_origin_x))		// Update the gorges position
			set_pev(id, pev_origin_y, pev_f(g_lifted[id], pev_origin_y))		// with the Lerks position
			set_pev(id, pev_origin_z, pev_f(g_lifted[id], pev_origin_z)-16)		// only 16 units lower

			set_pev(id, pev_velocity, {0.0,0.0,0.0})				// Disables fall damage
		}
	}

	return PLUGIN_CONTINUE
}

public switchOnOff(id, level, cid) {
	if (!cmd_access(id,level,cid,2))
		return PLUGIN_HANDLED

	new arg1[32]
	read_argv(1, arg1, 31)

	if (equal(arg1, "off")) {
		client_print(id, print_console, "[AMX] Lerk Lift disabled.")
		set_cvar_num("lerklift_enabled", 0)
		g_lldisabled = true

		for (new i=1; i<=g_MaxPlayers; ++i)
			if (g_lifted[i]) freeGorge(i)
	} else if (equal(arg1, "on")) {
		client_print(id, print_console, "[AMX] Lerk Lift enabled.")
		set_cvar_num("lerklift_enabled", 1)
		g_lldisabled = false
	} else
		client_print(id, print_console, "[AMX] Invalid option, valid options are 'off' or 'on'")

	return PLUGIN_CONTINUE
}

public client_disconnect(id) {
	if (g_lifting[id]) {		// Disconnected player was a Lerk carrying a Gorge
		freeGorge(g_lifting[id])
	} else if (g_lifted[id]) {	// Disconnected player was a Gorge being carried by a Lerk
		freeGorge(id)
	}
}

public client_putinserver(id) 
	set_task(20.0, "showUsage", id)

public showUsage(id) {
	if ( (g_lldisabled) || (pev(id, pev_team) != ALIEN) )
		return PLUGIN_HANDLED

	client_print(id, print_chat, "[AMX] Lerk Lift is enabled on this server. To disable Lerks from picking you up say /lloff ( /llon to enable )")

	return PLUGIN_HANDLED
}

public disableLL(id) {
	new CPstring[32]
	get_user_info(id, "cp", CPstring, 31)
	new CPvalue = str_to_num(CPstring)

	if (CPvalue & CP_LLNOLIFT) {
		client_print(id, print_chat, "[AMX] Lerk Lift is already disabled")
	} else {
		CPvalue |= CP_LLNOLIFT
		num_to_str(CPvalue, CPstring, 31)
		set_user_info(id, "cp", CPstring)
		client_print(id, print_chat, "[AMX] You can no longer be picked up by Lerks")
	}

	return PLUGIN_CONTINUE
}

public enableLL(id) {
	new CPstring[32]
	get_user_info(id, "cp", CPstring, 31)
	new CPvalue = str_to_num(CPstring)

	if (CPvalue & CP_LLNOLIFT) {
		CPvalue ^= CP_LLNOLIFT
		num_to_str(CPvalue, CPstring, 31)
		set_user_info(id, "cp", CPstring)
		client_print(id, print_chat, "[AMX] You can now be picked up by Lerks")
	} else {
		client_print(id, print_chat, "[AMX] Lerk Lift is already enabled")
	}

	return PLUGIN_CONTINUE
}

freeGorge(id) {
	if (is_user_connected(id) && is_user_alive(id)) {
		moveToFreeLocation(id)
		set_pev_i(id, pev_solid, SOLID_SLIDEBOX)
	}
	notifyStatus(id, g_lifted[id], 2)

	g_lifting[g_lifted[id]] = 0
	g_lifted[id] = 0
}

// Status:
// 1: Pickup Gorge
// 2: Release Gorge
notifyStatus(gorgeid, lerkid, status) {
	if ( is_user_connected(gorgeid) && is_user_alive(gorgeid) ) {
		if (is_user_connected(lerkid)) {
			new lerkname[32]
			get_user_name(lerkid,lerkname, 31)
			switch(status) {
				case 1: client_print(gorgeid, print_center, "You have been lifted by %s", lerkname)
				case 2: client_print(gorgeid, print_center, "You have detached from %s", lerkname)
			}
		} else client_print(gorgeid, print_center, "You have detached from the Lerk")
	}
	if ( is_user_connected(lerkid) && is_user_alive(lerkid) ) {
		if (is_user_connected(gorgeid)) {
			new gorgename[32]
			get_user_name(gorgeid, gorgename, 31)
			switch(status) {
				case 1: client_print(lerkid, print_center, "Picked up %s", gorgename)
				case 2: client_print(lerkid, print_center, "%s has been released", gorgename)
			}
		} else client_print(lerkid, print_center, "The gorge has been released")
	}
}

moveToFreeLocation(id) {
	new Float:origin[3], Float:new_origin[3], hullsize, distance

	hullsize = HULL_HEAD	// Only handling gorges here. See my unstuck to see how to handle other aliens.

	pev(id, pev_origin, origin)
	origin[2] += 16.0
	distance = START_DISTANCE

	while( distance < 1000 ) {	// 1000 is just incase of bugs, should never get anywhere near that
		for (new i = 0; i < MAX_ATTEMPTS; ++i) {
			new_origin[0] = random_float(origin[0]-distance,origin[0]+distance)
			new_origin[1] = random_float(origin[1]-distance,origin[1]+distance)
			new_origin[2] = random_float(origin[2]-distance,origin[2]+distance)

			if ( trace_hull(new_origin, hullsize, id) == 0 ) {
				set_origin(id, new_origin)
				return true
			}
		}
		distance += START_DISTANCE
	}
	return false
}
